import torch
import torch.nn as nn
from torch.nn.functional import cosine_similarity
import sys
sys.path.append("..")
import copy


class bgrl(nn.Module):
	"""
	https://github.com/nerdslab/bgrl/blob/4f38463683f21df517f55b3e88510146bc0dce57/bgrl/bgrl.py
	"""
	
	def __init__(self, args=None):
		"""
		dim: feature dimension (default: [batch_size, 512])
		m: moco momentum of updating key encoder (default: 0.999)
		T: softmax temperature (default: 0.07)
		"""
		super(bgrl, self).__init__()
		self.update_lr = args.lr
		self.predictor = args.predictor
		self.encoder_q = args.encoder_q
		self.encoder_k = args.encoder_k
		
		 
		for param in self.encoder_k.parameters():
			param.requires_grad = False
	
	@torch.no_grad()
	def update_target_network(self, mm):
		"""
		Performs a momentum update of the target network's weights.
		Args:
			mm (float): Momentum used in moving average update.
		"""
		assert 0.0 <= mm <= 1.0, "Momentum needs to be between 0.0 and 1.0, got %.5f" % mm
		for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
			param_k.data.mul_(mm).add_(param_q.data, alpha=1. - mm)
	
	@torch.no_grad()
	def _batch_shuffle_single_gpu(self, key):
		"""
		Batch shuffle, for making use of BatchNorm.
		*** Only support 1 gpu model. ***
		"""
		 
		batch_size_all = key.shape[0]
		 
		idx_shuffle = torch.randperm(batch_size_all)
		idx_shuffle = idx_shuffle.to(dtype=torch.long, device='cuda')
		 
		idx_unshuffle = torch.argsort(idx_shuffle)
		idx_unshuffle = idx_unshuffle.to(dtype=torch.long, device='cuda')
		
		key = key[idx_shuffle].to('cuda', non_blocking=True)
		return key, idx_unshuffle
		
	@torch.no_grad()
	def _batch_unshuffle_single_gpu(self, x, idx_unshuffle):
		"""
		Undo batch shuffle.
		*** Only support 1 gpu model. ***
		"""
		return x[idx_unshuffle]

	def byol_loss(self, q1, y2, q2, y1):
		return 2 - cosine_similarity(q1, y2.detach(), dim=-1).mean() - cosine_similarity(q2, y1.detach(), dim=-1).mean()

	def forward(self, g_0, features_0, g_1, features_1, to_fetch=None):
		"""
		Input:
			im_q: a batch of query inputs
			im_k: a batch of key inputs
		Output:
			logits, targets
		"""
		 
		q_0 = self.encoder_q(g_0, features_0, to_fetch)   
		q_1 = self.encoder_q(g_1, features_1, to_fetch)   

		q_0 = self.predictor(q_0.squeeze())
		q_1 = self.predictor(q_1.squeeze())
		
		 
		with torch.no_grad():   
			k_0 = self.encoder_k(g_0, features_0, to_fetch).detach()
			k_0, idx_unshuffle = self._batch_shuffle_single_gpu(k_0.squeeze())
			k_0 = self._batch_unshuffle_single_gpu(k_0, idx_unshuffle)
			
			k_1 = self.encoder_k(g_1, features_1, to_fetch).detach()
			k_1, idx_unshuffle = self._batch_shuffle_single_gpu(k_1.squeeze())
			k_1 = self._batch_unshuffle_single_gpu(k_1, idx_unshuffle)
		
		return self.byol_loss(q1=q_0, y2=k_1, q2=q_1, y1=k_0)


class bgrl_distill(bgrl):
	"""
	https://github.com/nerdslab/bgrl/blob/4f38463683f21df517f55b3e88510146bc0dce57/bgrl/bgrl.py
	"""
	
	def __init__(self, args=None):
		"""
		dim: feature dimension (default: [batch_size, 512])
		m: moco momentum of updating key encoder (default: 0.999)
		T: softmax temperature (default: 0.07)
		"""
		super(bgrl_distill, self).__init__(args=args)
		
	def forward(self, g_0, features_0, g_1, features_1, to_fetch=None):
		"""
		Input:
			im_q: a batch of query inputs
			im_k: a batch of key inputs
		Output:
			logits, targets
		"""
		 
		q_0 = self.encoder_q(g_0, features_0, to_fetch)   
		q_1 = self.encoder_q(g_1, features_1, to_fetch)   
		
		 
		with torch.no_grad():   
			k_0 = self.encoder_k(g_0, features_0, to_fetch).detach()
			k_0, idx_unshuffle = self._batch_shuffle_single_gpu(k_0.squeeze())
			k_0 = self._batch_unshuffle_single_gpu(k_0, idx_unshuffle)
			
			k_1 = self.encoder_k(g_1, features_1, to_fetch).detach()
			k_1, idx_unshuffle = self._batch_shuffle_single_gpu(k_1.squeeze())
			k_1 = self._batch_unshuffle_single_gpu(k_1, idx_unshuffle)
		
		return self.byol_loss(q1=q_0, y2=k_1, q2=q_1, y1=k_0)