import os
import sys
import time
import torch
import dgl
from utils import AverageMeter, ProgressMeter, adjust_learning_rate, adjust_moco_momentum

def train(train_loader, model, optimizer, scaler, console, epoch, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    learning_rates = AverageMeter('LR', ':.4e')
    losses = AverageMeter('Loss', ':.4e')

     
    model.train()

    end = time.time()
    train_loader=list(train_loader)
    iters_per_epoch = len(train_loader)
    moco_m = args.moco_m
    for i, (input_x_0, input_x_1) in enumerate(train_loader):
         
        data_time.update(time.time() - end)
        print('input_x_0.batch_size: ', input_x_0.batch_size, 'len(dgl.unbatch(input_x_0)): ', len(dgl.unbatch(input_x_0)))
         
         
         
        if args.moco_m_cos:
            moco_m = adjust_moco_momentum(epoch + i / iters_per_epoch, args)
         
         
         
         
         
        input_x_0 = input_x_0.to('cuda', non_blocking=True)
        input_x_1 = input_x_1.to('cuda', non_blocking=True)
         
         
         

         

         
        
        g_0, features_0, \
        g_1, features_1 = \
            input_x_0, input_x_0[:].ndata['h'], \
            input_x_1, input_x_1[:].ndata['h']

         
        with torch.cuda.amp.autocast(enabled=args.use_amp):
            loss = model(g_0, features_0, g_1, features_1, moco_m)

        losses.update(loss.item(), g_0.batch_size)

         
        optimizer.zero_grad()
        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()

         
        batch_time.update(time.time() - end)
        end = time.time()

        with open(args.outdir + '/val_loss_val.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.val) + '\n')
        with open(args.outdir + '/val_loss_avg.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.avg) + '\n')
        if i % args.print_freq == 0:
            console.write_log('Train epoch {0}: [{1}/{2}]\t'
            'Batch Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t''Data Time {data_time.val:.3f} ({data_time.avg:.3f})\t'
            'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
            .format(epoch, i, len(train_loader),
            batch_time=batch_time, data_time=data_time,
             loss=losses,))


def train_byol(train_loader, model, optimizer, scaler, console, epoch, args, lr_scheduler, mm_scheduler):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    learning_rates = AverageMeter('LR', ':.4e')
    losses = AverageMeter('Loss', ':.4e')
    
     
    model.train()
    
    end = time.time()
    train_loader = list(train_loader)
    iters_per_epoch = len(train_loader)
    for i, (input_x_0, input_x_1) in enumerate(train_loader):
    
        mm = 1 - mm_scheduler.get(i+epoch*len(train_loader))
        lr = lr_scheduler.get(i+epoch*len(train_loader))
        for g in optimizer.param_groups:
            g['lr'] = lr
    
         
        data_time.update(time.time() - end)

        input_x_0 = input_x_0.to('cuda', non_blocking=True)
        input_x_1 = input_x_1.to('cuda', non_blocking=True)
        
        
        g_0, features_0, \
        g_1, features_1 = \
            input_x_0, input_x_0[:].ndata['h'], \
            input_x_1, input_x_1[:].ndata['h']
        
         
        with torch.cuda.amp.autocast(enabled=args.use_amp):
            loss = model(g_0, features_0, g_1, features_1)
        
        losses.update(loss.item(), g_0.batch_size)
        
         
        optimizer.zero_grad()
        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()
        model.update_target_network(mm)
         
        batch_time.update(time.time() - end)
        end = time.time()
        
        with open(args.outdir + '/val_loss_val.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.val) + '\n')
        with open(args.outdir + '/val_loss_avg.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.avg) + '\n')
        with open(args.outdir + '/lr.txt', 'a', encoding='utf-8') as f:
            f.write(str(lr) + '\n')
        with open(args.outdir + '/mm.txt', 'a', encoding='utf-8') as f:
            f.write(str(mm) + '\n')
        if i % args.print_freq == 0:
            console.write_log('Ep {0}: [{1}/{2}]\t'
                              'Batch Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t''Data Time {data_time.val:.3f} ({data_time.avg:.3f})\t'
                              'Loss {loss.val:.4f} ({loss.avg:.4f})\tlr {lr}\tmm {mm}'
                              .format(epoch, i, len(train_loader),
                                      batch_time=batch_time, data_time=data_time,
                                      loss=losses, lr=lr, mm=mm))


def train_byol_l2l(train_loader, model, optimizer, scaler, console, epoch, args, lr_scheduler, mm_scheduler):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    learning_rates = AverageMeter('LR', ':.4e')
    losses = AverageMeter('Loss', ':.4e')
    
     
    model.train()
    
    end = time.time()
    iters_per_epoch = len(train_loader)
    for i, (input_x_0, input_x_1, input_x_center) in enumerate(train_loader):
        
        mm = 1 - mm_scheduler.get(i + epoch * len(train_loader))
        lr = lr_scheduler.get(i + epoch * len(train_loader))
        for g in optimizer.param_groups:
            g['lr'] = lr
        
         
        data_time.update(time.time() - end)
        
        input_x_0 = input_x_0.to('cuda', non_blocking=True)
        input_x_1 = input_x_1.to('cuda', non_blocking=True)
        input_x_center = input_x_center.to('cuda', non_blocking=True)
        assert input_x_center is not None

        g_0, features_0, \
        g_1, features_1 = \
            input_x_0, input_x_0[:].ndata['h'], \
            input_x_1, input_x_1[:].ndata['h']
        
         
        with torch.cuda.amp.autocast(enabled=args.use_amp):
            loss = model(g_0, features_0, g_1, features_1, input_x_center)
        
        losses.update(loss.item(), g_0.batch_size)
        
         
        optimizer.zero_grad()
        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()
        model.update_target_network(mm)
         
        batch_time.update(time.time() - end)
        end = time.time()
        
        with open(args.outdir + '/val_loss_val.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.val) + '\n')
        with open(args.outdir + '/val_loss_avg.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.avg) + '\n')
        with open(args.outdir + '/lr.txt', 'a', encoding='utf-8') as f:
            f.write(str(lr) + '\n')
        with open(args.outdir + '/mm.txt', 'a', encoding='utf-8') as f:
            f.write(str(mm) + '\n')
        if i % args.print_freq == 0:
            console.write_log('Ep {0}: [{1}/{2}]\t'
                              'Batch Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t''Data Time {data_time.val:.3f} ({data_time.avg:.3f})\t'
                              'Loss {loss.val:.4f} ({loss.avg:.4f})\tlr {lr}\tmm {mm}'
                              .format(epoch, i, len(train_loader),
                                      batch_time=batch_time, data_time=data_time,
                                      loss=losses, lr=lr, mm=mm))


def train_byol_distill(train_loader, model, optimizer, scaler, console, epoch, args, lr_scheduler, mm_scheduler):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    learning_rates = AverageMeter('LR', ':.4e')
    losses = AverageMeter('Loss', ':.4e')
    
     
    model.train()
    
    end = time.time()
    train_loader = list(train_loader)
    iters_per_epoch = len(train_loader)
    for i, (input_x_0, input_x_1, input_x_center) in enumerate(train_loader):
        
         
        lr = lr_scheduler.get(i + epoch * len(train_loader))
        for g in optimizer.param_groups:
            g['lr'] = lr
        
         
        data_time.update(time.time() - end)
        
        input_x_0 = input_x_0.to('cuda', non_blocking=True)
        input_x_1 = input_x_1.to('cuda', non_blocking=True)
        input_x_center = input_x_center.to('cuda', non_blocking=True)
        assert input_x_center is not None
        
        g_0, features_0, \
        g_1, features_1 = \
            input_x_0, input_x_0[:].ndata['h'], \
            input_x_1, input_x_1[:].ndata['h']
        
         
        with torch.cuda.amp.autocast(enabled=args.use_amp):
            loss = model(g_0, features_0, g_1, features_1, input_x_center)
        
        losses.update(loss.item(), g_0.batch_size)
        
         
        optimizer.zero_grad()
        scaler.scale(loss).backward()
        scaler.step(optimizer)
        scaler.update()
         
         
        batch_time.update(time.time() - end)
        end = time.time()
        
        with open(args.outdir + '/val_loss_val.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.val) + '\n')
        with open(args.outdir + '/val_loss_avg.txt', 'a', encoding='utf-8') as f:
            f.write(str(losses.avg) + '\n')
        with open(args.outdir + '/lr.txt', 'a', encoding='utf-8') as f:
            f.write(str(lr) + '\n')
         
         
        if i % args.print_freq == 0:
            console.write_log('Ep {0}: [{1}/{2}]\t'
                              'Batch Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t''Data Time {data_time.val:.3f} ({data_time.avg:.3f})\t'
                              'Loss {loss.val:.4f} ({loss.avg:.4f})\tlr {lr}'
                              .format(epoch, i, len(train_loader),
                                      batch_time=batch_time, data_time=data_time,
                                      loss=losses, lr=lr)) 