import numpy as np
import torch
import random
from collections import defaultdict
from torch.utils.data import Dataset
import dgl
from torch.utils.data import DataLoader


class FewShotBatchSampler:
	def __init__(self, dataset_targets, N_way, K_shot, include_query=False, shuffle=True, shuffle_once=False):
		"""
		Inputs:
			dataset_targets - PyTorch tensor of the labels of the data elements.
			N_way - Number of classes to sample per batch.
			K_shot - Number of examples to sample per class in the batch.
			include_query - If True, returns batch of size N_way*K_shot*2, which
							can be split into support and query set. Simplifies
							the implementation of sampling the same classes but
							distinct examples for support and query set.
			shuffle - If True, examples and classes are newly shuffled in each
					  iteration (for training)
			shuffle_once - If True, examples and classes are shuffled once in
						   the beginning, but kept constant across iterations
						   (for validation)
		"""
		super().__init__()
		self.dataset_targets = dataset_targets
		self.N_way = N_way
		self.K_shot = K_shot
		self.shuffle = shuffle
		self.include_query = include_query
		if self.include_query:
			self.K_shot *= 2
		self.batch_size = self.N_way * self.K_shot   
		
		 
		self.classes = torch.unique(self.dataset_targets).tolist()
		self.num_classes = len(self.classes)
		self.indices_per_class = {}
		self.batches_per_class = {}   
		for c in self.classes:
			self.indices_per_class[c] = torch.where(self.dataset_targets == c)[0]
			self.batches_per_class[c] = self.indices_per_class[c].shape[0] // self.K_shot
		
		 
		self.iterations = sum(self.batches_per_class.values()) // self.N_way
		self.class_list = [c for c in self.classes for _ in range(self.batches_per_class[c])]
		if shuffle_once or self.shuffle:
			self.shuffle_data()
		else:
			 
			sort_idxs = [
				i + p * self.num_classes for i, c in enumerate(self.classes) for p in range(self.batches_per_class[c])
			]
			self.class_list = np.array(self.class_list)[np.argsort(sort_idxs)].tolist()
	
	def shuffle_data(self):
		 
		for c in self.classes:
			perm = torch.randperm(self.indices_per_class[c].shape[0])
			self.indices_per_class[c] = self.indices_per_class[c][perm]
		 
		 
		 
		random.shuffle(self.class_list)
	
	def __iter__(self):
		 
		if self.shuffle:
			self.shuffle_data()
		
		 
		start_index = defaultdict(int)
		for it in range(self.iterations):
			class_batch = self.class_list[it * self.N_way: (it + 1) * self.N_way]   
			index_batch = []
			for c in class_batch:   
				index_batch.extend(self.indices_per_class[c][start_index[c]: start_index[c] + self.K_shot])
				start_index[c] += self.K_shot
			if self.include_query:   
				index_batch = index_batch[::2] + index_batch[1::2]
			yield index_batch
	
	def __len__(self):
		return self.iterations


class fsl_dataset(Dataset):
	def __init__(self, root='/home/ name/data/graph/TRIANGLES/val_set.bin'):
		super(fsl_dataset, self).__init__()
		self.graph, self.label = dgl.data.utils.load_graphs(root)
		 
		self.targets = self.label['glabel']   
	
	 
	 
	 
	def __getitem__(self, index):
		 
		 
		 
		return self.graph[index], self.targets[index].item()
	
	def __len__(self):
		 
		return len(self.graph)


def collate_fsl(samples):
	 
	graph, label = map(list, zip(*samples))
	graph = dgl.batch(graph)
	return graph, label


def list_of_groups(init_list, childern_list_len):
	'''
	init_list为初始化的列表，childern_list_len初始化列表中的几个数据组成一个小列表
	:param init_list:
	:param childern_list_len:
	:return:
	'''
	list_of_group = zip(*(iter(init_list),) * childern_list_len)
	end_list = [list(i) for i in list_of_group]
	count = len(init_list) % childern_list_len
	end_list.append(init_list[-count:]) if count != 0 else end_list
	return end_list


if __name__ == '__main__':
	 
	train_set = fsl_dataset(root='/home/ name/data/graph/TRIANGLES/train_set.bin')
	sampler = FewShotBatchSampler(train_set.targets, include_query=True, N_way=3, K_shot=10, shuffle=True)
	data_loader = DataLoader(train_set, num_workers=2, pin_memory=True, persistent_workers=True, collate_fn=collate_fsl,
	                         batch_sampler=sampler)
	
	for step, (graph, label) in enumerate(data_loader):
		print(step)
		graph = dgl.unbatch(graph)
		x_spt = graph[0:len(label) // 2]
		x_spt = list_of_groups(x_spt, 10)   
		print('x_spt: ', x_spt)
		x_qry = graph[len(label) // 2:-1]
		y_spt = label[0:len(label) // 2]
		y_spt = list_of_groups(y_spt, 10)
		print('x_spt: ', y_spt)
		y_qry = label[len(label) // 2:-1]
	 
	 
	 
	 
	 
	 
	 
