import pickle
import numpy as np
import torch
import random
from collections import defaultdict
from torch.utils.data import Dataset
import dgl


class FewShotBatchSampler:
	def __init__(self, dataset_targets, N_way, K_shot, include_query=False, shuffle=True, shuffle_once=False):
		"""
		Inputs:
			dataset_targets - PyTorch tensor of the labels of the data elements.
			N_way - Number of classes to sample per batch.
			K_shot - Number of examples to sample per class in the batch.
			include_query - If True, returns batch of size N_way*K_shot*2, which
							can be split into support and query set. Simplifies
							the implementation of sampling the same classes but
							distinct examples for support and query set.
			shuffle - If True, examples and classes are newly shuffled in each
					  iteration (for training)
			shuffle_once - If True, examples and classes are shuffled once in
						   the beginning, but kept constant across iterations
						   (for validation)
		"""
		super().__init__()
		self.dataset_targets = dataset_targets
		self.N_way = N_way
		self.K_shot = K_shot
		self.shuffle = shuffle
		self.include_query = include_query
		if self.include_query:
			self.K_shot *= 2
		self.batch_size = self.N_way * self.K_shot   
		
		 
		self.classes = torch.unique(self.dataset_targets).tolist()
		self.num_classes = len(self.classes)
		self.indices_per_class = {}
		self.batches_per_class = {}   
		for c in self.classes:
			self.indices_per_class[c] = torch.where(self.dataset_targets == c)[0]
			self.batches_per_class[c] = self.indices_per_class[c].shape[0] // self.K_shot
		
		 
		self.iterations = sum(self.batches_per_class.values()) // self.N_way
		self.class_list = [c for c in self.classes for _ in range(self.batches_per_class[c])]
		if shuffle_once or self.shuffle:
			self.shuffle_data()
		else:
			 
			sort_idxs = [
				i + p * self.num_classes for i, c in enumerate(self.classes) for p in range(self.batches_per_class[c])
			]
			self.class_list = np.array(self.class_list)[np.argsort(sort_idxs)].tolist()
	
	def shuffle_data(self):
		 
		for c in self.classes:
			perm = torch.randperm(self.indices_per_class[c].shape[0])
			self.indices_per_class[c] = self.indices_per_class[c][perm]
		 
		 
		 
		random.shuffle(self.class_list)
	
	def __iter__(self):
		 
		if self.shuffle:
			self.shuffle_data()
		
		 
		start_index = defaultdict(int)
		for it in range(self.iterations):
			class_batch = self.class_list[it * self.N_way: (it + 1) * self.N_way]   
			index_batch = []
			for c in class_batch:   
				index_batch.extend(self.indices_per_class[c][start_index[c]: start_index[c] + self.K_shot])
				start_index[c] += self.K_shot
			if self.include_query:   
				index_batch = index_batch[::2] + index_batch[1::2]
			yield index_batch
	
	def __len__(self):
		return self.iterations


class fsl_dataset(Dataset):
	def __init__(self, root='/home/ name/data/graph/TRIANGLES/train_set.bin'):
		super(fsl_dataset, self).__init__()
		self.graph = dgl.data.utils.load_graphs(root)
	
	def __getitem__(self, index):
		return self.graph[index], self.label[index]
	
if __name__ == '__main__':
	def load_pickle(file_name):
		with open(file_name, 'rb') as f:
			data = pickle.load(f)
			return data
	
	saved_set = load_pickle('/home/ name/data/graph/TRIANGLES/TRIANGLES_val_set.pickle')
	label2graphs, graph2nodes, graph2edges = saved_set["label2graphs"], saved_set["graph2nodes"], saved_set[
		"graph2edges"]
	
	print(len(graph2nodes))
	print('graph2edges:', len(graph2edges))
	graph_list = []
	for i in label2graphs:
		for graph_id in label2graphs[i]:
			graph_list.append(graph_id)
	 
	
	node_attribures_path = '/home/ name/data/graph/TRIANGLES/TRIANGLES_node_attributes.pickle'
	attrs = load_pickle(node_attribures_path)
	attrs = list(map(float, attrs))
	global_attributes = attrs
	node_attribures = global_attributes
	print('node_attribures:', node_attribures)
	edge_index = []
	graph_indicator = []
	node_attr = []
	node_number = 0
	labels = torch.from_numpy(np.load('/home/ name/data/graph/TRIANGLES/val_labels.npy'))
	graph_labels = {"glabel": labels}
	print('labels:', labels)
	graph_set = []
	for index, gid in enumerate(graph_list):
		node_number = 0
		nodes = graph2nodes[gid]
		new_nodes = list(range(node_number, node_number + len(nodes)))
		node_number = node_number + len(nodes)
		node2new_number = dict(zip(nodes, new_nodes))
		
		node_attr_gid = np.array([node_attribures[node] for node in nodes]).reshape(len(nodes), -1)
		node_attr.append(node_attr_gid)
		edge_index_gid = [[node2new_number[edge[0]], node2new_number[edge[1]]] for edge in graph2edges[gid]]
		edge_index.extend(edge_index_gid)
		graph_indicator.extend([index] * len(nodes))
		
		edge_index_gid = np.array(edge_index_gid).transpose(1,0)
		u = torch.from_numpy(edge_index_gid[0])
		v = torch.from_numpy(edge_index_gid[1])
		graph_dgl = dgl.graph((u, v))
		graph_dgl.ndata['h'] = torch.from_numpy(node_attr_gid)
		
		graph_set.append(graph_dgl)
		print(index)
		 
		 
		 
		 
	dgl.data.utils.save_graphs('/home/ name/data/graph/TRIANGLES/val_set.bin',graph_set, graph_labels)
	
	node_attr = np.vstack(node_attr)
	print('len(edge_index):', len(edge_index))
	print('len(node_attr):', len(node_attr))

	graph_label = {}
	
	for label in label2graphs:
		for graph in label2graphs[label]:
			graph_label[graph] = label
	 
	
	 
	 
	print(len(graph_label))
	 
	dataset_targets = []
	for graph in graph2nodes:
		 
		dataset_targets.append(graph_label[graph])
	print(len(dataset_targets))
	 
