import torch
import torch.nn as nn
import sys
sys.path.append("..")
from gmeta.learner import gat_net


class MoCo(nn.Module):
	"""
	Build a MoCo model with: a query encoder, a key encoder, and a queue
	https://arxiv.org/abs/1911.05722
	"""
	
	def __init__(self, dim=128, K=131072, m=0.999, T=0.07, args=None):
		"""
		dim: feature dimension (default: [batch_size, 512])
		K: queue size; number of negative keys (default: 65536)
		m: moco momentum of updating key encoder (default: 0.999)
		T: softmax temperature (default: 0.07)
		"""
		super(MoCo, self).__init__()
		
		self.dim = dim
		self.K = K
		self.m = m
		self.T = T
		
		 
		 
		self.update_lr = args.lr
		self.encoder_q = gat_net(hidden_mlp_dim=args.hidden_mlp_dim, out_mlp_dim=dim,
                        n_layers=args.n_layers, in_dim=args.n_feats, n_hidden=args.n_hidden,
                        n_classes=args.n_classes, heads=args.heads, activation=args.activation,
                        feat_drop=args.in_drop, attn_drop=args.attn_drop, negative_slope=args.negative_slope, residual=args.residual, norm=args.norm)
		self.encoder_k = gat_net(hidden_mlp_dim=args.hidden_mlp_dim, out_mlp_dim=dim,
                        n_layers=args.n_layers,
                        in_dim=args.n_feats, n_hidden=args.n_hidden, n_classes=args.n_classes,
                        heads=args.heads, activation=args.activation,
                        feat_drop=args.in_drop, attn_drop=args.attn_drop, negative_slope=args.negative_slope, residual=args.residual, norm=args.norm)
		print(self.encoder_q)
		for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
			param_k.data.copy_(param_q.data)   
			param_k.requires_grad = False   
		
		 
		self.register_buffer("queue", torch.randn(self.dim, K))
		self.queue = nn.functional.normalize(self.queue, dim=0)
		
		self.register_buffer("queue_ptr", torch.zeros(1, dtype=torch.long))
	
	@torch.no_grad()
	def _momentum_update_key_encoder(self, moco_m):
		""" Momentum update of the key encoder """
		self.m = moco_m
		for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
			param_k.data = param_k.data * self.m + param_q.data * (1. - self.m)
	
	@torch.no_grad()
	def _dequeue_and_enqueue(self, keys):
		batch_size = keys.shape[0]
		
		ptr = int(self.queue_ptr)
		assert self.K % batch_size == 0   
		
		 
		self.queue[..., ptr:ptr + batch_size] = keys.T
		ptr = (ptr + batch_size) % self.K   
		
		self.queue_ptr[0] = ptr
	
	@torch.no_grad()
	def _batch_shuffle_single_gpu(self, key):
		"""
		Batch shuffle, for making use of BatchNorm.
		*** Only support 1 gpu model. ***
		"""
		 
		batch_size_all = key.shape[0]
		 
		 
		idx_shuffle = torch.randperm(batch_size_all)
		idx_shuffle = idx_shuffle.to(dtype=torch.long, device='cuda')
		 
		idx_unshuffle = torch.argsort(idx_shuffle)
		idx_unshuffle = idx_unshuffle.to(dtype=torch.long, device='cuda')
		
		key = key[idx_shuffle].to('cuda', non_blocking=True)
		
		return key, idx_unshuffle
	
		
	@torch.no_grad()
	def _batch_unshuffle_single_gpu(self, x, idx_unshuffle):
		"""
		Undo batch shuffle.
		*** Only support 1 gpu model. ***
		"""
		return x[idx_unshuffle]
	
	def contrastive_loss(self, logits, labels):
		return nn.CrossEntropyLoss()(logits, labels) * (2 * self.T)
	
	def forward(self, g_0, features_0, g_1, features_1, moco_m):
		"""
		Input:
			im_q: a batch of query inputs
			im_k: a batch of key inputs
		Output:
			logits, targets
		"""
		 
		 
		q_0 = self.encoder_q(g_0, features_0)   
		q_1 = self.encoder_q(g_1, features_1)   
		
		 
		q_0 = self.encoder_q.fc(q_0.squeeze())
		 
		q_0 = self.encoder_q.predictor(q_0)
		q_1 = self.encoder_q.predictor(self.encoder_q.fc(q_1.squeeze()))
		
		q_0 = nn.functional.normalize(q_0, dim=1)
		q_1 = nn.functional.normalize(q_1, dim=1)
		
		 
		with torch.no_grad():   
			self._momentum_update_key_encoder(moco_m)   
			 
			k_0 = self.encoder_k(g_0, features_0)
			k_0, idx_unshuffle = self._batch_shuffle_single_gpu(k_0.squeeze())
			k_0 = self.encoder_k.predictor(self.encoder_k.fc(k_0))
			k_0 = nn.functional.normalize(k_0, dim=1)   
			
			 
			k_0 = self._batch_unshuffle_single_gpu(k_0, idx_unshuffle)
			
			 
			k_1 = self.encoder_k(g_1, features_1)
			k_1, idx_unshuffle = self._batch_shuffle_single_gpu(k_1.squeeze())
			k_1 = self.encoder_k.predictor(self.encoder_k.fc(k_1))
			k_1 = nn.functional.normalize(k_1, dim=1)   
			
			 
			k_1 = self._batch_unshuffle_single_gpu(k_1, idx_unshuffle)
		
		 
		 
		 
		l_pos_0 = torch.einsum('nc,nc->n', [q_0, k_1]).unsqueeze(-1)
		l_pos_1 = torch.einsum('nc,nc->n', [q_1, k_0]).unsqueeze(-1)
		 
		 
		 
		 
		l_neg_0 = torch.einsum('nc,ck->nk', [q_0, self.queue.clone().detach()])
		l_neg_1 = torch.einsum('nc,ck->nk', [q_1, self.queue.clone().detach()])
		
		 
		logits_0 = torch.cat([l_pos_0, l_neg_0], dim=1)
		logits_1 = torch.cat([l_pos_1, l_neg_1], dim=1)
		 
		logits_0 /= self.T
		logits_1 /= self.T
		
		 
		labels_0 = torch.zeros(logits_0.shape[0], dtype=torch.long).cuda()
		labels_1 = torch.zeros(logits_1.shape[0], dtype=torch.long).cuda()
		
		 
		self._dequeue_and_enqueue(k_0)
		self._dequeue_and_enqueue(k_1)
		
		return self.contrastive_loss(logits_0, labels_0) + self.contrastive_loss(logits_1, labels_1)
