import  torch, os
import sys
import torch.multiprocessing
torch.multiprocessing.set_sharing_strategy('file_system')

sys.path.append('..')

curPath = os.path.abspath(os.path.dirname(__file__))
rootPath = os.path.split(curPath)[0]
sys.path.append(rootPath)

from    cl_data_processing import cl_subgraphs, l2l_subgraphs, cl_collate, l2l_collate
from    torch.utils.data import DataLoader
import  random, sys, pickle
import  argparse
from dgl.data.utils import load_graphs, save_graphs
import numpy as np
from gcl.bgrl import bgrl_distill
from gmeta.gnn_zoo import GCN, MLP_Predictor
import time
from contrast_trainer import train_byol_distill
from utils import save_checkpoint, console_log, CosineDecayScheduler, load_pretrained
import torch.nn.functional as F


def main():
    torch.manual_seed(0)
    torch.cuda.manual_seed_all(0)
    np.random.seed(0)
    random.seed(0)
    
    root = args.data_dir
    feat = np.load(root + 'features.npy', allow_pickle=True)
    args.n_feats = feat.shape[1]

    dgl_graph, _label_dict = load_graphs(root+'/graph_dgl.bin')

    if args.task_setup == 'Disjoint':
        with open(root + 'label.pkl', 'rb') as f:
            info = pickle.load(f)
    elif args.task_setup == 'Shared':
        if args.task_mode == 'True':
            root = root + '/task' + str(args.task_n) + '/'
        with open(root + 'label.pkl', 'rb') as f:
            info = pickle.load(f)

    if len(feat.shape) == 2:
         
        feat = [feat]
    
    device = torch.device('cuda')
    args.n_classes = len(list(set(info)))
    
    prepare_dataset_start = time.time()
    if args.l2l:
        db_train = l2l_subgraphs(root, 'train', info, args=args, adjs=dgl_graph, h=args.h, feat=feat)
    else:
        db_train = cl_subgraphs(root, 'train', info, args=args, adjs=dgl_graph, h=args.h, feat=feat)
    print('Time on dataset preparing:', str(time.time() - prepare_dataset_start)[:5])
    args.n_classes = db_train.num_labels_classes

    args.encoder_q = GCN(in_feats=args.n_feats,
                         n_hidden=args.n_hidden,
                         n_classes=args.n_classes,
                         n_layers=args.n_layers,
                         activation=args.activation,
                         dropout=args.dropout,
                         batchnorm=args.batchnorm,
                         ssl=True,
                         residual=args.residual,
                         l2l=args.l2l)
    args.encoder_k = GCN(in_feats=args.n_feats,
                         n_hidden=args.n_hidden,
                         n_classes=args.n_classes,
                         n_layers=args.n_layers,
                         activation=args.activation,
                         dropout=args.dropout,
                         batchnorm=args.batchnorm,
                         ssl=True,
                         residual=args.residual,
                         l2l=args.l2l)
    load_pretrained(args.encoder_k, args.ckpt_path, console, True)
    args.pred_dim = args.n_hidden // 2
    args.predictor = None
    
    model = bgrl_distill(args=args)
    model = model.to(device)
    print('bgrl model: ', model)
    
    if args.l2l:
        train_dataloader = DataLoader(db_train, args.batchsz, shuffle=True, num_workers=args.num_workers, pin_memory=True,
                                      collate_fn=l2l_collate, persistent_workers=True, drop_last=True)
    else:
        train_dataloader = DataLoader(db_train, args.batchsz, shuffle=True, num_workers=args.num_workers, pin_memory=True,
                                  collate_fn=cl_collate, persistent_workers=True, drop_last=True)
    train_dataloader = list(train_dataloader)

    optimizer = torch.optim.AdamW(model.parameters(), lr=args.lr)

    
    console.write_log(str(args))
    print('------ Start Training ------')
    s_start = time.time()
    
    args.steps = len(train_dataloader) * args.epochs
    args.lr_warmup_steps = len(train_dataloader) * args.warmup_epochs
    
    lr_scheduler = CosineDecayScheduler(args.lr, args.lr_warmup_steps, args.steps)
    mm_scheduler = CosineDecayScheduler(1 - args.mm, 0, args.steps)
    
    scaler = torch.cuda.amp.GradScaler(enabled=args.use_amp)
    for epoch in range(args.start_epoch, args.epochs):
         
        train_byol_distill(train_dataloader, model, optimizer, scaler, console, epoch, args, lr_scheduler, mm_scheduler)
        save_checkpoint({
            'epoch': epoch,
            'args': args,
            'state_dict': model.state_dict(),
            'optimizer': optimizer.state_dict(),
            'scaler': scaler.state_dict(),
        }, is_best=False, filename=args.outdir+'/ckpt.pth.tar')

    print('Total Time:', str(time.time() - s_start)[:5])
    
    
if __name__ == '__main__':
    
    argparser = argparse.ArgumentParser()
    argparser.add_argument('--outdir', default='outputs/', type=str, help='path to location where to save outputs. Empty for current working dir')
    argparser.add_argument('--exp_name', default='distill', type=str, help='name under outputs/')
    argparser.add_argument('--use_amp', action='store_true', help='whether use amp')
    argparser.add_argument('--start_epoch', type=int, help='start epoch number', default=0)
    argparser.add_argument('--warmup_epochs', type=int, help='warmup epochs number, if ==0 then no warmup', default=5)
    argparser.add_argument('--epochs', type=int, help='total epoch number', default=100)
    argparser.add_argument('--lr', type=float, help='bgrl learning rate', default=0.05)
    argparser.add_argument('--input_dim', type=int, help='input feature dim', default=1)
    argparser.add_argument('--hidden_dim', type=int, help='hidden dim', default=128)
    argparser.add_argument('--attention_size', type=int, help='dim of attention_size', default=32)
    argparser.add_argument('--data_dir', default='/home/ name/data/arxiv/', type=str, required=False, help="The input dataset dir.")
    argparser.add_argument('--task_setup', default='Disjoint', type=str, required=False, help="Select from Disjoint or Shared Setup. For Disjoint-Label, single/multiple graphs are both considered.")
    argparser.add_argument('--num_workers', default=8, type=int, required=False, help="num of workers")
    argparser.add_argument('--batchsz', default=8192, type=int, required=False, help="batch size")
    argparser.add_argument("--ckpt_path", default='/home/ name/research/G-Meta/outputs/bgrl/3_l2l/bs8192-warmup5-lr0.05-m0.99-layer1/ckpt.pth.tar',
                           type=str, required=False, help="cl pretrained ckpt dir.")
    argparser.add_argument('--link_pred_mode', default='False', type=str, required=False, help="For Link Prediction")
    argparser.add_argument('--h', default=3, type=int, required=False, help="neighborhood size")
    argparser.add_argument('--sample_nodes', type=int, help='sample nodes if above this number of nodes', default=1000)
    argparser.add_argument('--mm', default=0.999, type=float, help='bgrl momentum of updating momentum encoder (default: 0.99)')
    
    argparser.add_argument('--n_layers', type=int, default=1, help="number of hidden layers")
    argparser.add_argument('--n_feats', type=int, help="number of features[1]")
    argparser.add_argument('--n_hidden', type=int, default=256, help="number of hidden units")
    argparser.add_argument('--n_classes', type=int, default=128, help="number of classes, equal to GNN output dimension")
    argparser.add_argument('--pred_dim', type=int, default=128, help="number of classes, equal to GNN output dimension")
    argparser.add_argument('--in_drop', type=float, default=0, help="input feature dropout, especially for GAT")
    argparser.add_argument('--attn_drop', type=float, default=0, help="attention dropout, especially for GAT")
    argparser.add_argument('--negative_slope', type=float, default=0.2, help="the negative slope of leaky relu, especially for GAT")
    argparser.add_argument('--residual', type=bool, default=False, help="use residual for gnn")
    argparser.add_argument('--num_heads', type=int, default=8, help="number of hidden attention heads")
    argparser.add_argument('--num_out_heads', type=int, default=1, help="number of output attention heads")
    argparser.add_argument('--heads', type=int, default=1, help="number of all heads in GAT")
    argparser.add_argument('--activation', help="activations (F.elu) in GAT")
    argparser.add_argument('-p', '--print_freq', default=10, type=int, metavar='N', help='print frequency (default: 10)')
    argparser.add_argument('--dropout', type=float, default=0.0, help="dropout rate for GNN")
    argparser.add_argument('--batchnorm', type=bool, default=True, help="use batchnorm for GNN")
    argparser.add_argument('--l2l', type=bool, default=True, help="use local to local contrast")
    argparser.add_argument('--norm', action='store_true', default=False, help="use F.normalize in GAT output")


    args = argparser.parse_args()
    args.activation = F.leaky_relu   
    args.heads = ([args.num_heads] * args.n_layers) + [args.num_out_heads]
    args.outdir = os.path.join(args.outdir, args.exp_name,
                               'bs{bs}-warmup{warmup_epochs}-lr{lr}-m{bgrl_m}-layer{n_layers}'.format(
                                   bs=args.batchsz, bgrl_m=args.mm, warmup_epochs=args.warmup_epochs,
                                    n_layers=args.n_layers, lr=args.lr))

    if os.path.exists(args.outdir) is None:
        os.makedirs(os.path.join(args.outdir), exist_ok=True)
    print(str(args))
    console = console_log(args.outdir)
    main()
