import os
import torch
import math
import shutil
import numpy as np


class SubsetRandomSampler(torch.utils.data.Sampler):
	r"""Samples elements randomly from a given list of indices, without replacement.

	Arguments:
		indices (sequence): a sequence of indices
	"""
	
	def __init__(self, indices):
		self.epoch = 0
		self.indices = indices
	
	def __iter__(self):
		return (self.indices[i] for i in torch.randperm(len(self.indices)))
	
	def __len__(self):
		return len(self.indices)
	
	def set_epoch(self, epoch):
		self.epoch = epoch


def save_checkpoint(state, is_best, filename='ckpt.pth.tar'):
	torch.save(state, filename)
	if is_best:
		shutil.copyfile(filename, 'best.pth.tar')


def load_pretrained(model, ckpt_path, console, strict=False):
	console.write_log('loading ssl_ckpt... ', ckpt_path)
	model_dict = model.state_dict()
	checkpoint = torch.load(ckpt_path)
	state_dict = checkpoint['state_dict']
	 
	for k in list(state_dict.keys()):
		 
		if k.startswith('encoder_q'):
			 
			state_dict[k[len("encoder_q."):]] = state_dict[k]
			if k[len("encoder_q."):].startswith('fc') or k[len("encoder_q."):].startswith('predictor'):
				del state_dict[k[len("encoder_q."):]]
		 
		del state_dict[k]
	for k in list(state_dict.keys()):
		print(k)
	msg = model.load_state_dict(state_dict, strict=strict)
	console.write_log('messages in Decoder:', msg)
	console.write_log('missing_keys in Decoder:', msg.missing_keys)
	console.write_log('loading checkpoint successfully')


class AverageMeter(object):
	"""Computes and stores the average and current value"""
	
	def __init__(self, name, fmt=':f'):
		self.name = name
		self.fmt = fmt
		self.reset()
	
	def reset(self):
		self.val = 0
		self.avg = 0
		self.sum = 0
		self.count = 0
	
	def update(self, val, n=1):
		self.val = val
		self.sum += val * n
		self.count += n
		self.avg = self.sum / self.count
	
	def __str__(self):
		fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
		return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
	def __init__(self, num_batches, meters, prefix=""):
		self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
		self.meters = meters
		self.prefix = prefix
	
	def display(self, batch):
		entries = [self.prefix + self.batch_fmtstr.format(batch)]
		entries += [str(meter) for meter in self.meters]
		print('\t'.join(entries))
	
	def _get_batch_fmtstr(self, num_batches):
		num_digits = len(str(num_batches // 1))
		fmt = '{:' + str(num_digits) + 'd}'
		return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
	"""Decays the learning rate with half-cycle cosine after warmup"""
	if epoch < args.warmup_epochs:
		lr = args.lr * epoch / args.warmup_epochs
	else:
		lr = args.lr * 0.5 * (
				1. + math.cos(math.pi * (epoch - args.warmup_epochs) / (args.epochs - args.warmup_epochs)))
	for param_group in optimizer.param_groups:
		param_group['lr'] = lr
	return lr


def adjust_moco_momentum(epoch, args):
	"""Adjust moco momentum based on current epoch"""
	m = 1. - 0.5 * (1. + math.cos(math.pi * epoch / args.epochs)) * (1. - args.moco_m)
	return m


class console_log(object):
	def __init__(self, logs_path='./'):
		self.logs_path = logs_path
		if not os.path.exists(self.logs_path):
			os.makedirs(self.logs_path, exist_ok=True)
	
	def write_log(self, log_str=None, should_print=True, prefix='console', end='\n'):
		with open(os.path.join(self.logs_path, '%s.log' % prefix), 'a') as fout:
			fout.write(log_str + end)
			fout.flush()
		if should_print:
			print(log_str)


class CosineDecayScheduler:
	def __init__(self, max_val, warmup_steps, total_steps):
		self.max_val = max_val
		self.warmup_steps = warmup_steps
		self.total_steps = total_steps
	
	def get(self, step):
		if step < self.warmup_steps:
			return self.max_val * step / self.warmup_steps
		elif self.warmup_steps <= step <= self.total_steps:
			return self.max_val * (1 + np.cos((step - self.warmup_steps) * np.pi /
			                                  (self.total_steps - self.warmup_steps))) / 2
		else:
			raise ValueError('Step ({}) > total number of steps ({}).'.format(step, self.total_steps))


if __name__ == '__main__':
	support_y_relative = [10, 20, 30, 40]
	support_y = [0, 0, 0, 0]
	l = 3
	idx = 5
	print(support_y == 0)
	print(support_y_relative[True])
 
 
