"""GCN using DGL nn package
reference code : https://github.com/dmlc/dgl/tree/master/examples/pytorch
References:
- Semi-Supervised Classification with Graph Convolutional Networks
- Paper: https://arxiv.org/abs/1609.02907
- Code: https://github.com/tkipf/gcn
"""
import torch
import torch.nn as nn
from dgl.nn.pytorch import GraphConv
import torch.nn.functional as F
from dgl.nn.pytorch.conv import GINConv
from dgl.nn.pytorch.glob import SumPooling, AvgPooling, MaxPooling
from dgllife.model.gnn.gcn import GCNLayer
from learn2learn.utils import clone_module, update_module

class GCN(nn.Module):
    def __init__(self,
                 in_feats,
                 n_hidden,
                 n_classes,
                 n_layers,
                 activation,
                 dropout,
                 batchnorm=True, ssl=True, residual=False, LinkPred_mode=False, l2l=False):
        super(GCN, self).__init__()
        self.layers = nn.ModuleList()
         
        self.layers.append(GCNLayer(in_feats, n_hidden, activation=activation, dropout=dropout))
         
        for i in range(n_layers - 1):
            self.layers.append(GCNLayer(n_hidden, n_hidden, activation=activation, batchnorm=batchnorm,
                                        residual=residual, dropout=dropout))
         
        self.layers.append(GCNLayer(n_hidden, n_hidden//2, batchnorm=batchnorm, residual=residual, dropout=dropout, activation=activation,))
         
        self.ssl = ssl
        self.l2l = l2l
        if self.ssl:
            if l2l:
                self.readout = nn.Identity()   
            else:
                self.readout = nn.Identity()
        else:
            self.readout = nn.Linear(n_hidden//2, n_classes, bias=False) 
        self.LinkPred_mode = LinkPred_mode
        
    def forward(self, g, features, to_fetch=None):
        if self.ssl:
            if self.l2l:
                return self._l2l_forward(g, features, to_fetch)
            else:
                return self._cl_forward(g, features)
        else:
            return self._ft_forward(g, features, to_fetch)
        
    def _cl_forward(self, g, features):
        h = features
         
        for i, layer in enumerate(self.layers):
            g.ndata['h'] = h
             
            h = layer(g, g.ndata['h'])
        g.ndata['h'] = h
        h = self.readout(g, h)
        return h

    def _l2l_forward(self, g, features, to_fetch):
        h = features
        g.ndata['h'] = h
        for i, layer in enumerate(self.layers):
            h = layer(g, h)
            g.ndata['h'] = h
    
        num_nodes_ = g.batch_num_nodes().tolist()
        temp = [0] + num_nodes_
        offset = torch.cumsum(torch.LongTensor(temp), dim=0)[:-1].to('cuda')
    
        if self.LinkPred_mode:
            h1 = h[to_fetch[:, 0] + offset]
            h2 = h[to_fetch[:, 1] + offset]
            h = torch.cat((h1, h2), 1)
        else:
            h = h[to_fetch + offset]
        h = self.readout(h)
        return h
    
    def _ft_forward(self, g, features, to_fetch):
        h = features
        g.ndata['h'] = h
        for i, layer in enumerate(self.layers):
            h = layer(g, h)
            g.ndata['h'] = h

        num_nodes_ = g.batch_num_nodes().tolist()
        temp = [0] + num_nodes_
        offset = torch.cumsum(torch.LongTensor(temp), dim=0)[:-1].to('cuda')

        if self.LinkPred_mode:
            h1 = h[to_fetch[:, 0] + offset]
            h2 = h[to_fetch[:, 1] + offset]
            h = torch.cat((h1, h2), 1)
        else:
            h = h[to_fetch + offset]
        h = self.readout(h)
        return h


"""
How Powerful are Graph Neural Networks
https://arxiv.org/abs/1810.00826
https://openreview.net/forum?id=ryGs6iA5Km
Author's implementation: https://github.com/weihua916/powerful-gnns
"""
class ApplyNodeFunc(nn.Module):
    """Update the node feature hv with MLP, BN and ReLU."""
    def __init__(self, mlp):
        super(ApplyNodeFunc, self).__init__()
        self.mlp = mlp
        self.bn = nn.BatchNorm1d(self.mlp.output_dim)

    def forward(self, h):
        h = self.mlp(h)
        h = self.bn(h)
        h = F.relu(h)
        return h


class MLP_GIN(nn.Module):
    """MLP with linear output"""
    def __init__(self, n_layers, input_dim, hidden_dim, output_dim):
        """MLP layers construction

        Paramters
        ---------
        n_layers: int
            The number of linear layers
        input_dim: int
            The dimensionality of input features
        hidden_dim: int
            The dimensionality of hidden units at ALL layers
        output_dim: int
            The number of classes for prediction

        """
        super(MLP_GIN, self).__init__()
        self.linear_or_not = True   
        self.n_layers = n_layers
        self.output_dim = output_dim

        if n_layers < 1:
            raise ValueError("number of layers should be positive!")
        elif n_layers == 1:
             
            self.linear = nn.Linear(input_dim, output_dim)
        else:
             
            self.linear_or_not = False
            self.linears = torch.nn.ModuleList()
            self.batch_norms = torch.nn.ModuleList()

            self.linears.append(nn.Linear(input_dim, hidden_dim))
            for layer in range(n_layers - 2):
                self.linears.append(nn.Linear(hidden_dim, hidden_dim))
            self.linears.append(nn.Linear(hidden_dim, output_dim))

            for layer in range(n_layers - 1):
                self.batch_norms.append(nn.BatchNorm1d((hidden_dim)))

    def forward(self, x):
        if self.linear_or_not:
             
            return self.linear(x)
        else:
             
            h = x
            for i in range(self.n_layers - 1):
                h = F.relu(self.batch_norms[i](self.linears[i](h)))
            return self.linears[-1](h)


class GIN(nn.Module):
    """GIN model"""
    def __init__(self, n_layers, num_mlp_layers, input_dim, hidden_dim,
                 output_dim, final_dropout, learn_eps, graph_pooling_type,
                 neighbor_pooling_type):
        """model parameters setting

        Paramters
        ---------
        n_layers: int
            The number of linear layers in the neural network
        num_mlp_layers: int
            The number of linear layers in mlps
        input_dim: int
            The dimensionality of input features
        hidden_dim: int
            The dimensionality of hidden units at ALL layers
        output_dim: int
            The number of classes for prediction
        final_dropout: float
            dropout ratio on the final linear layer
        learn_eps: boolean
            If True, learn epsilon to distinguish center nodes from neighbors
            If False, aggregate neighbors and center nodes altogether.
        neighbor_pooling_type: str
            how to aggregate neighbors (sum, mean, or max)
        graph_pooling_type: str
            how to aggregate entire nodes in a graph (sum, mean or max)

        """
        super(GIN, self).__init__()
        self.n_layers = n_layers
        self.learn_eps = learn_eps

         
        self.ginlayers = torch.nn.ModuleList()
        self.batch_norms = torch.nn.ModuleList()

        for layer in range(self.n_layers - 1):
            if layer == 0:
                mlp = MLP_GIN(num_mlp_layers, input_dim, hidden_dim, hidden_dim)
            else:
                mlp = MLP_GIN(num_mlp_layers, hidden_dim, hidden_dim, hidden_dim)

            self.ginlayers.append(
                GINConv(ApplyNodeFunc(mlp), neighbor_pooling_type, 0, self.learn_eps))
            self.batch_norms.append(nn.BatchNorm1d(hidden_dim))

         
         
        self.linears_prediction = torch.nn.ModuleList()

        for layer in range(n_layers):
            if layer == 0:
                self.linears_prediction.append(
                    nn.Linear(input_dim, output_dim))
            else:
                self.linears_prediction.append(
                    nn.Linear(hidden_dim, output_dim))

        self.drop = nn.Dropout(final_dropout)

        if graph_pooling_type == 'sum':
            self.pool = SumPooling()
        elif graph_pooling_type == 'mean':
            self.pool = AvgPooling()
        elif graph_pooling_type == 'max':
            self.pool = MaxPooling()
        else:
            raise NotImplementedError

    def forward(self, g, h):
         
        hidden_rep = [h]

        for i in range(self.n_layers - 1):
            h = self.ginlayers[i](g, h)
            h = self.batch_norms[i](h)
            h = F.relu(h)
            hidden_rep.append(h)

        score_over_layer = 0

         
        for i, h in enumerate(hidden_rep):
            pooled_h = self.pool(g, h)
            score_over_layer += self.drop(self.linears_prediction[i](pooled_h))

        return score_over_layer

class MLP_Predictor(nn.Module):
    r"""MLP used for predictor. The MLP has one hidden layer.
    Args:
        input_size (int): Size of input features.
        output_size (int): Size of output features.
        hidden_size (int, optional): Size of hidden layer. (default: :obj:`4096`).
    """
    def __init__(self, input_size, output_size, hidden_size=128):
        super().__init__()
        hidden_size = input_size * 2
        self.net = nn.Sequential(
            nn.Linear(input_size, hidden_size, bias=False),
            nn.PReLU(1),
            nn.Linear(hidden_size, output_size, bias=True)
        )
        self.reset_parameters()

    def forward(self, x):
        return self.net(x)

    def reset_parameters(self):
         
        for m in self.modules():
            if isinstance(m, nn.Linear):
                m.reset_parameters()