 
 

 

import torch
from torch import nn
from torch import optim
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
import json
import logging
import os, copy
from scipy.stats import entropy
from torch.distributions import Categorical

torch.multiprocessing.set_sharing_strategy('file_system')
import sys

sys.path.append("..")
import torch.multiprocessing

curPath = os.path.abspath(os.path.dirname(__file__))
rootPath = os.path.split(curPath)[0]
sys.path.append(rootPath)


def calculate_regularization(sampled_x, Phi, reduced_axes=None, device=None):
	""" Calculate the variance that is used for Interpreter

	Args:
		sample_x (list of torch.FloatTensor):
			A list of sampled input embeddings $x$, each $x$ is of shape ``[length, dimension]``. All the $x$s can have different length,
			but should have the same dimension. Sampled number should be higher to get a good estimation.
		Phi (function):
			The $Phi$ we studied. A function whose input is x (element in the first parameter) and returns a hidden state (of type
			``torch.FloatTensor``, of any shape)
		reduced_axes (list of ints, Optional):
			The axes that is variable in Phi (e.g., the sentence length axis). We will reduce these axes by mean along them.

	Returns:
		torch.FloatTensor: The regularization term calculated

	"""
	sample_num = len(sampled_x)
	sample_s = []
	for n in range(sample_num):
		x = sampled_x[n]
		if device is not None:
			x = x.to(device)
		s = Phi(x)
		if reduced_axes is not None:
			for axis in reduced_axes:
				assert axis < len(s.shape)
				s = s.mean(dim=axis, keepdim=True)
		sample_s.append(s.tolist())
	sample_s = np.array(sample_s)
	return np.std(sample_s, axis=0)


class Interpreter(nn.Module):
	""" Interpreter for interpret one instance.

	It will minimize the loss in Eqn.(7):

		$L(sigma) = (||Phi(embed + epsilon) - Phi(embed)||_2^2) // (regularization^2) - rate * log(sigma)$

	In our implementation, we use reparameterization trick to represent epsilon ~ N(0, sigma^2 I), i.e. epsilon = scale * ratio * noise.
	Where noise ~ N(0, 1), scale is a hyper-parameter that controls the maximum value of sigma^2, and ratio in (0, 1) is the learnable parameter.

	"""
	
	def __init__(self, x, Phi=None, scale=0.5, rate=0.1, regularization=None, words=None, model=None, c_x=None):
		""" Initialize an interpreter class.

		Args:
			x (torch.FloatTensor): Of shape ``[length, dimension]``.
				The $x$ we studied. i.e. The input word embeddings.
			Phi (function):
				The $Phi$ we studied. A function whose input is x (the first parameter) and returns a hidden state (of type ``torch.FloatTensor``, of any shape)
			scale (float):
				The maximum size of sigma. A hyper-parameter in reparameterization trick. The recommended value is 10 * Std[word_embedding_weight],
				where word_embedding_weight is the word embedding weight in the model interpreted. Larger scale will give more salient result, Default: 0.5.
			rate (float):
				A hyper-parameter that balance the MLE Loss and Maximum Entropy Loss. Larger rate will result in larger information loss. Default: 0.1.
			regularization (Torch.FloatTensor or np.ndarray):
				The regularization term, should be of the same shape as (or broadcastable to) the output of Phi. If None is given, method will use the output to
				regularize itself. Default: None.
			words (List[Str]):
				The input sentence, used for visualizing. If None is given, method will not show the words.

		"""
		super(Interpreter, self).__init__()
		self.s = x.ndata['h'].size(0)
		self.d = x.ndata['h'].size(1)
		self.ratio = nn.Parameter(torch.randn(self.s, 1), requires_grad=True)
		self.model = model
		self.c_x = c_x
		
		self.scale = scale
		self.rate = rate
		self.x = x
		self.Phi = Phi
		
		self.regular = regularization
		if self.regular is not None:
			self.regular = nn.Parameter(torch.tensor(self.regular).to(x.ndata['h']), requires_grad=False)
		self.words = words
		if self.words is not None:
			assert self.s == len(words), 'the length of x should be of the same with the lengh of words'
	
	def forward(self):
		""" Calculate loss:

			$L(sigma) = (||Phi(embed + epsilon) - Phi(embed)||_2^2) // (regularization^2) - rate * log(sigma)$

		Returns:
			torch.FloatTensor: a scalar, the target loss.

		"""
		ratios = torch.sigmoid(self.ratio)   
		x = self.x
		x.ndata['h'] = self.x.ndata['h'] + 0.   
		x_tilde = x
		
		 
		s = self.model(x, x.ndata['h'], self.c_x)
		
		 
		x_tilde.ndata['h'] = x_tilde.ndata['h'] + ratios * torch.randn(self.s, self.d).to(
			x.device) * self.scale   
		 
		s_tilde = self.model(x_tilde, x_tilde.ndata['h'], self.c_x)
		loss = (s_tilde - s) ** 2
		if self.regular is not None:
			loss = torch.mean(loss / self.regular ** 2)
		else:
			loss = torch.mean(loss) / torch.mean(s ** 2)
		
		return loss - torch.mean(torch.log(ratios)) * self.rate
	
	def optimize(self, iteration=5000, lr=0.01, show_progress=False, x_graph=None, c_x=None):
		""" Optimize the loss function
		Args:
			iteration (int): Total optimizing iteration
			lr (float): Learning rate
			show_progress (bool): Whether to show the learn progress
		"""
		minLoss = None
		state_dict = None
		optimizer = optim.Adam(self.parameters(), lr=lr)
		self.train()
		func = (lambda x: x) if not show_progress else tqdm
		x_list = []
		for _ in range(iteration):
			x_list.append(copy.deepcopy(x_graph))
		
		for idx, _ in enumerate(range(iteration)):
			print(idx)
			optimizer.zero_grad()
			
			 
			ratios = torch.sigmoid(self.ratio)   
			x_list[idx].ndata['h'] = x_list[idx].ndata['h'] + 0.   
			x_tilde = copy.deepcopy(x_list[idx])
			
			 
			with torch.no_grad():
				s = self.model(x_list[idx], x_list[idx].ndata['h'], c_x)
			
			 
			x_tilde.ndata['h'] = x_tilde.ndata['h'] + ratios * torch.randn(self.s, self.d).to(
				x_list[idx].device) * self.scale   
			 
			s_tilde = self.model(x_tilde, x_tilde.ndata['h'], c_x)
			loss = (s_tilde - s) ** 2
			if self.regular is not None:
				loss = torch.mean(loss / self.regular ** 2)
			else:
				loss = torch.mean(loss) / torch.mean(s ** 2)
			loss_total = loss - torch.mean(torch.log(ratios)) * self.rate
			 
			
			loss_total.backward()
			optimizer.step()
			 
			if minLoss is None or minLoss > loss.item():
				state_dict = {k: self.state_dict()[k] + 0. for k in self.state_dict().keys()}
				minLoss = loss.item()
			del loss
		
		self.eval()
		self.load_state_dict(state_dict)
		del x_list
		del self.model
		self.model = None
		
	def get_sigma(self):
		""" Calculate and return the sigma

		Returns:
			np.ndarray: of shape ``[seqLen]``, the ``sigma``.

		"""
		ratios = torch.sigmoid(self.ratio)   
		return ratios.detach().cpu().numpy()[:, 0] * self.scale
	
	def get_info_loss(self):
		""" Calculate and return the sigma

		Returns:
			np.ndarray: of shape ``[seqLen]``, the ``sigma``.

		"""
		ratios = torch.sigmoid(self.ratio)   
		return ratios.sum()
	
	def visualize(self):
		""" Visualize the information loss of every word.
		"""
		sigma_ = self.get_sigma()
		_, ax = plt.subplots()
		ax.imshow([sigma_], cmap='GnBu_r')
		ax.set_xticks(range(self.s))
		 
		ax.set_yticks([0])
		ax.set_yticklabels([''])
		plt.tight_layout()
		plt.show()


class interpreter_dataset(nn.Module):
	""" Interpreter for interpret one instance.

	It will minimize the loss in Eqn.(7):

		$L(sigma) = (||Phi(embed + epsilon) - Phi(embed)||_2^2) // (regularization^2) - rate * log(sigma)$

	In our implementation, we use reparameterization trick to represent epsilon ~ N(0, sigma^2 I), i.e. epsilon = scale * ratio * noise.
	Where noise ~ N(0, 1), scale is a hyper-parameter that controls the maximum value of sigma^2, and ratio in (0, 1) is the learnable parameter.

	"""
	
	def __init__(self, scale=0.5, rate=0.1):
		""" Initialize an interpreter class.

		Args:
			x (torch.FloatTensor): Of shape ``[length, dimension]``.
				The $x$ we studied. i.e. The input word embeddings.
			Phi (function):
				The $Phi$ we studied. A function whose input is x (the first parameter) and returns a hidden state (of type ``torch.FloatTensor``, of any shape)
			scale (float):
				The maximum size of sigma. A hyper-parameter in reparameterization trick. The recommended value is 10 * Std[word_embedding_weight],
				where word_embedding_weight is the word embedding weight in the model interpreted. Larger scale will give more salient result, Default: 0.5.
			rate (float):
				A hyper-parameter that balance the MLE Loss and Maximum Entropy Loss. Larger rate will result in larger information loss. Default: 0.1.
			regularization (Torch.FloatTensor or np.ndarray):
				The regularization term, should be of the same shape as (or broadcastable to) the output of Phi. If None is given, method will use the output to
				regularize itself. Default: None.
			words (List[Str]):
				The input sentence, used for visualizing. If None is given, method will not show the words.
		"""
		super(interpreter_dataset, self).__init__()
		self.scale = scale
		self.rate = rate
		self.regular = None
		self.ratio_list = []
		
	def optimize(self, iteration=5000, lr=0.01, show_progress=False, dataloader=None, model=None, feat=None):
		""" Optimize the loss function
		Args:
			iteration (int): Total optimizing iteration
			lr (float): Learning rate
			show_progress (bool): Whether to show the learn progress
		"""
		model.eval()
		print('model:', model)
		 
		for x_spt, y_spt, x_qry, y_qry, c_spt, c_qry, n_spt, n_qry, g_spt, g_qry in tqdm(dataloader):
			for i in range(len(x_qry)):
				feat_spt = torch.Tensor(
					np.vstack(([feat[g_spt[i][j]][np.array(x)] for j, x in enumerate(n_spt[i])]))).to('cuda')
				x_spt[i] = x_spt[i].to('cuda')
				c_spt[i] = c_spt[i].to('cuda')
				x_spt[i].ndata['h'] = feat_spt
				x_spt[i] = x_spt[i].to('cuda')

				minLoss = None
				state_dict = None
				func = (lambda x: x) if not show_progress else tqdm
				x_list = []
				s_graph = x_spt[i].ndata['h'].size(0)
				dim_graph = x_spt[i].ndata['h'].size(1)
				self.ratio = nn.Parameter(torch.randn(s_graph, 1), requires_grad=True) 
				self.to('cuda')
				optimizer = optim.Adam(self.parameters(), lr=lr)
				
				self.train()
				
				for idx, _ in enumerate(range(iteration)):
					x_list.append(copy.deepcopy(x_spt[i]))
					optimizer.zero_grad()
					 
					ratios = torch.sigmoid(self.ratio)   
					x_list[idx].ndata['h'] = x_list[idx].ndata['h'] + 0.   
					x_tilde = copy.deepcopy(x_list[idx])
					
					with torch.no_grad():
						s = model(x_list[idx], x_list[idx].ndata['h'], c_spt[i])
					
					x_tilde.ndata['h'] = x_tilde.ndata['h'] + ratios * torch.randn(s_graph, dim_graph).to(
						x_list[idx].device) * self.scale   
					s_tilde = model(x_tilde, x_tilde.ndata['h'], c_spt[i])
					loss = (s_tilde - s) ** 2
					if self.regular is not None:
						loss = torch.mean(loss / self.regular ** 2)
					else:
						loss = torch.mean(loss) / torch.mean(s ** 2)
					loss_total = loss - torch.mean(torch.log(ratios)) * self.rate
					loss_total.backward()
					optimizer.step()
					 
					
					if minLoss is None or minLoss > loss.item():
						state_dict = {k: self.state_dict()[k] + 0. for k in self.state_dict().keys()}
						minLoss = loss.item()
					del loss
				self.ratio_list.append(self.ratio)
				self.load_state_dict(state_dict)
				del x_list
		self.eval()
		return self.get_log_sigma()
		
	def get_info_loss(self):
		ratios_sum_list = []
		for ratio in self.ratio_list:
			ratio = torch.sigmoid(ratio)   
			ratios_sum_list.append(ratio.sum().item())
		return sum(ratios_sum_list)/len(ratios_sum_list)

	def get_entropy(self):
		ratio_entropy_sum_list = []
		for ratio in self.ratio_list:
			ratio = torch.sigmoid(ratio)   
			ratio_entropy = Categorical(probs=ratio).entropy()
			ratio_entropy_sum_list.append(sum(ratio_entropy.tolist()))
		return sum(ratio_entropy_sum_list) / len(ratio_entropy_sum_list)
	
	def get_log_sigma(self):
		log_ratio_sum_list = []
		for ratio in self.ratio_list:
			 
			ratio = torch.sigmoid(ratio)   
			ratio = torch.squeeze(ratio)
			log_ratio = torch.log(ratio)   
			 

			 
			log_ratio_sum_list.append(sum(log_ratio.tolist()))
		return sum(log_ratio_sum_list) / len(log_ratio_sum_list)
	


if __name__ == '__main__':
	text = "rare bird has more than enough charm to make it memorable."
